// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"
	"path"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/version"
)

// Logger is a logger automatically configured by command line options
var Logger xbus.Logger

var (
	cfgFile string
	v       = viper.New()
)

// RootCmd represents the base command when called without any subcommands
var RootCmd = &cobra.Command{
	Use:   "xbus-http",
	Short: "Xbus http gateway",
	Long: `Xbus http gateway utility

xbus-http provides rest gateway to xbus.`,
	Run: version.Cmd,
}

// Execute adds all child commands to the root command sets flags appropriately.
// This is called by main.main(). It only needs to happen once to the rootCmd.
func Execute() {
	if err := RootCmd.Execute(); err != nil {
		fmt.Println(err)
		os.Exit(-1)
	}
}

// GetClient returns a XBusClient connected (if relevant) to the bus using the cli options
func GetClient(nocert bool) (*xbus.Client, error) {
	tlsSaver, err := xbus.LoadTLSOptionFiles(v, v.ConfigFileUsed(), Logger)
	if err != nil {
		return nil, err
	}

	configDir := path.Dir(v.ConfigFileUsed())
	opts := xbus.COptionsFromViper(v, api.Account_GATEWAY, configDir, Logger)

	opts.TLS.TLSSaver = tlsSaver

	if nocert {
		opts.TLS.Certificate = nil
	}

	// Connect!
	c := xbus.NewClient(opts, Logger)

	if opts.ConnOptions.Url != "" {
		if err := c.Connect(); err != nil {
			Logger.Errorf("Error connecting to the NATS server: %s", err)
			return nil, err
		}
	}
	return c, nil
}

func init() {
	cobra.OnInitialize(initConfig)

	flags := RootCmd.PersistentFlags()
	version.InitFlags(flags)

	flags.StringVar(&cfgFile, "config", "", "config file; default is $HOME/.config/xbus/xbus-http.yaml)")

	flags.BoolP("debug", "d", false, "Enable debug mode")
	flags.Bool("trace", false, "Enable trace mode")

	v.BindPFlag("logging.debug", flags.Lookup("debug"))
	v.BindPFlag("logging.trace", flags.Lookup("trace"))

	xbus.DefineAndBindPFlags(v, flags)
}

// initConfig reads in config file and ENV variables if set.
// it also setup the logger
func initConfig() {
	if cfgFile != "" { // enable ability to specify config file via flag
		v.SetConfigFile(cfgFile)
	} else {
		v.SetConfigName("xbus-http") // name of config file (without extension)
	}

	v.SetEnvPrefix("XBUS_HTTP")
	v.AddConfigPath("$HOME/.config/xbus")
	v.AddConfigPath("/etc/xbus")
	v.AutomaticEnv() // read in environment variables that match

	Logger = xbus.GetLogger(&xbus.LoggerOptions{
		Debug: v.GetBool("logging.debug"),
		Trace: v.GetBool("logging.trace"),
	})
	// If a config file is found, read it in.
	if err := v.ReadInConfig(); err == nil {
		Logger.Debugf("Using config file: %s", v.ConfigFileUsed())
	}
}
