package pipeline

import (
	"fmt"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// setEqual compare two slices of strings and return true if they contain
// the same elements no matter the order
func stringSetEqual(s1, s2 []string) bool {
	if len(s1) != len(s2) {
		return false
	}
	for _, v1 := range s1 {
		found := false
		for _, v2 := range s2 {
			if v1 == v2 {
				found = true
				break
			}
		}
		if !found {
			return false
		}
	}
	return true
}

// NodeType is the type of Node
type NodeType int

const (
	// EmitterNode is the type of emitter nodes
	EmitterNode NodeType = iota
	// ConsumerNode is the type of consumer nodes
	ConsumerNode
	// WorkerNode is the type of worker nodes
	WorkerNode
)

// NewPipeline creates a graph
func NewPipeline(pipeline storage.Pipeline) Pipeline {
	return Pipeline{
		Pipeline: pipeline,
	}
}

// Pipeline is a single xbus pipeline
type Pipeline struct {
	storage.Pipeline `yaml:",inline"`
}

// GetDestinations returns the inputs on which to send envelope coming from given outputs
func (p Pipeline) GetDestinations(output storage.NodeOutputRef) ([]storage.NodeInputRef, error) {
	var dests []storage.NodeInputRef
	for _, edge := range p.Graph.Edges {
		if output == edge.From {
			dests = append(dests, edge.To)
		}
	}
	return dests, nil
}

// Match returns the source input corresponding to the emitter & eventtypes
// and true if it exists, empty value and false if not
func (p Pipeline) Match(
	emitter api.Actor, eventtypes []string,
) (storage.NodeOutputRef, bool) {
	for _, node := range p.Graph.Nodes {
		if node.Type == storage.EmitterNode &&
			stringSetEqual(node.SourceMatch.EventTypes, eventtypes) {
			return storage.NodeOutputRef{
				NodeID: node.ID,
				Output: api.DefaultOutput,
			}, true
		}
	}
	return storage.NodeOutputRef{}, false
}

// GetNode returns a node from its ID
func (p Pipeline) GetNode(id string) (storage.Node, error) {
	for _, node := range p.Graph.Nodes {
		if node.ID == id {
			return node, nil
		}
	}
	return storage.Node{}, fmt.Errorf("No such node: %s", id)
}
