package pipeline

import (
	"regexp"
	"strings"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/storage"
)

const idRegexpS = "^[a-zA-Z0-9_\\-]+$"

var idRegexp *regexp.Regexp

// A Path is a list of vertices
type Path []storage.Edge

// PathList is a list of Path
type PathList []Path

// Contains returns true if the Path contains the given node
func (p Path) Contains(id string) int {
	for i, edge := range p {
		if edge.From.NodeID == id {
			return i
		}
		if edge.To.NodeID == id {
			return i + 1
		}
	}
	return -1
}

// Clone returns a copy of the path with a capacity of +1
func (p Path) Clone() Path {
	np := make(Path, len(p), len(p)+1)
	copy(p, np)
	return np
}

// compareCycles returns true if 2 paths contains the same nodes.
// it does _not_ enforce that the path are indeed cycles
func compareCycles(p1, p2 Path) bool {
	if len(p1) != len(p2) {
		return false
	}
	edges := map[storage.Edge]bool{}
	for _, edge := range p1 {
		edges[edge] = false
	}
	for _, edge := range p2 {
		_, ok := edges[edge]
		if !ok {
			return false
		}
		edges[edge] = true
	}
	for _, ok := range edges {
		if !ok {
			return false
		}
	}
	return true
}

// ContainsCycle returns true if the given cycle is already
// in the path list
func (l PathList) ContainsCycle(cycle Path) bool {
	for _, c := range l {
		if compareCycles(cycle, c) {
			return true
		}
	}
	return false
}

// Messages returns the errors as a list of messages
func (l ValidationErrorList) Messages() []string {
	var strList = make([]string, 0, len(l))
	for _, e := range l {
		strList = append(strList, e.Error())
	}
	return strList
}

func (l ValidationErrorList) Error() string {
	return strings.Join(l.Messages(), "\n")
}

// NewValidator crates a Validator
func NewValidator(graph storage.Graph) Validator {
	return Validator{Graph: graph}
}

// Validator validates a graph consistency
type Validator struct {
	Graph storage.Graph

	edgesFrom map[string][]storage.Edge
	edgesTo   map[string][]storage.Edge
}

func (v *Validator) initMaps() {
	v.edgesFrom = make(map[string][]storage.Edge)
	v.edgesTo = make(map[string][]storage.Edge)
	for _, edge := range v.Graph.Edges {
		v.edgesFrom[edge.From.NodeID] = append(v.edgesFrom[edge.From.NodeID], edge)
		v.edgesTo[edge.To.NodeID] = append(v.edgesTo[edge.From.NodeID], edge)
	}
}

// findCycles finds all cycles starting from the given node
// It does so by doing a depth-first search and identifies all
// the paths that leads back to any node of the path
func (v *Validator) findCycles(nodeID string) PathList {
	s := PathList{}
	cycles := PathList{}

	for _, edge := range v.edgesFrom[nodeID] {
		if edge.To.NodeID == nodeID {
			cycles = append(cycles, Path{edge})
		} else {
			s = append(s, Path{edge})
		}
	}
	for len(s) != 0 {
		// Pop the last inserted path
		path := s[0]
		s = s[1:]

		tailNodeID := path[len(path)-1].To.NodeID

		for _, edge := range v.edgesFrom[tailNodeID] {
			if i := path.Contains(edge.To.NodeID); i != -1 {
				cycle := append(path[i:].Clone(), edge)
				if !cycles.ContainsCycle(cycle) {
					cycles = append(cycles, cycle)
				}
			} else {
				s = append(PathList{append(path.Clone(), edge)}, s...)
			}
		}
	}
	return cycles
}

// checkForCycles find and returns any cycle in the graph
func (v *Validator) checkForCycles() PathList {
	var cycles PathList

	for _, node := range v.Graph.Nodes {
		if edges, ok := v.edgesTo[node.ID]; !ok || len(edges) == 0 {
			for _, cycle := range v.findCycles(node.ID) {
				if !cycles.ContainsCycle(cycle) {
					cycles = append(cycles, cycle)
				}
			}
		}
	}
	return cycles
}

// isConnected returns true if the graph is connected.
// Does a depth-first exploration from any node (following edges in both ways)
// and verifies it reaches all the nodes of the graph
func (v *Validator) isConnected() bool {
	if len(v.Graph.Nodes) == 0 {
		return true
	}
	s := []string{v.Graph.Nodes[0].ID}
	visited := map[string]bool{s[0]: true}

	for len(s) != 0 {
		id := s[len(s)-1]
		s = s[:len(s)-1]
		for _, edge := range v.edgesFrom[id] {
			if !visited[edge.To.NodeID] {
				visited[edge.To.NodeID] = true
				s = append(s, edge.To.NodeID)
			}
		}
		for _, edge := range v.edgesTo[id] {
			if !visited[edge.From.NodeID] {
				visited[edge.From.NodeID] = true
				s = append(s, edge.From.NodeID)
			}
		}
	}
	if len(visited) > len(v.Graph.Nodes) {
		panic("Visited more nodes than the graph contains")
	}
	if len(visited) < len(v.Graph.Nodes) {
		return false
	}
	return true
}

// checkSourceMatch ensure a node source match consistency
func (v *Validator) checkSourceMatch(node storage.Node, errors *ValidationErrorList) {

	if node.SourceMatch == nil {
		errors.Add(MissingSourceMatch(node.ID))
		return
	}

	for _, event := range node.SourceMatch.EventTypes {
		if !tools.ValidateEventType(event) {
			errors.Add(SourceMatchInvalidEventType(node.ID, event))
		}
	}
}

// Validate enforce the graph is consistent.
func (v *Validator) Validate() (errors ValidationErrorList) {
	v.initMaps()

	// There one and exactly one emitter, and at least one consumer
	var foundEmitter, foundConsumer bool

	// make sure all edges reference declared input/output, and that
	// every input/output is connected to something
	allInputs := map[storage.NodeInputRef]bool{}
	allOutputs := map[storage.NodeOutputRef]bool{}

	// make sure all IDs are valid and there are no duplicates
	ids := make(map[string]int)

	for _, node := range v.Graph.Nodes {
		ids[node.ID] = ids[node.ID] + 1
		if ids[node.ID] == 1 {
			if !idRegexp.Match([]byte(node.ID)) {
				errors.Add(InvalidNodeID(node.ID))
			}
		} else if ids[node.ID] == 2 {
			errors.Add(DuplicateNodeIDError(node.ID))
		}
		if node.Type == storage.EmitterNode {
			if foundEmitter {
				errors.Add(ExtraEmitterError(node.ID))
			}
			foundEmitter = true

			v.checkSourceMatch(node, &errors)

			if len(node.Outputs) == 0 {
				errors.Add(EmitterHasNoOutput(node.ID))
			}
			if len(node.Outputs) > 1 {
				errors.Add(EmitterHasMultipleOutput(node.ID))
			}
			if len(node.Inputs) != 0 {
				errors.Add(EmitterHasInput(node.ID))
			}
		} else if node.Type == storage.ConsumerNode {
			foundConsumer = true

			if len(node.Inputs) == 0 {
				errors.Add(ConsumerHasNoInput(node.ID))
			}
			if len(node.Inputs) > 1 {
				errors.Add(ConsumerHasMultipleInput(node.ID))
			}
			if len(node.Outputs) != 0 {
				errors.Add(ConsumerHasOutput(node.ID))
			}
			if len(node.Roles) == 0 {
				errors.Add(ConsumerMustHave1Role(node.ID))
			}
		}
		for _, input := range node.Inputs {
			allInputs[storage.NodeInputRef{NodeID: node.ID, Input: input}] = false
		}
		for _, output := range node.Outputs {
			allOutputs[storage.NodeOutputRef{NodeID: node.ID, Output: output}] = false
		}

		// Enforce roles & broadcast are properly used (they currently have
		// some big restriction in the process implementation)
		if len(node.Roles) > 1 {
			errors.Add(TooManyRolesError(node.ID))
		}
		if len(node.Roles) >= 1 && !node.RoleBroadcast {
			errors.Add(RoleBroadcastUnsetError(node.ID))
		}

		// Check Source nodes are emitter and sink nodes are consumer
		if len(v.edgesFrom[node.ID]) != 0 &&
			len(v.edgesTo[node.ID]) == 0 &&
			node.Type != storage.EmitterNode {
			errors.Add(SourceIsNotEmitterError(node.ID))
		}
		if len(v.edgesTo[node.ID]) != 0 &&
			len(v.edgesFrom[node.ID]) == 0 &&
			node.Type != storage.ConsumerNode {
			errors.Add(SinkIsNotConsumerError(node.ID))
		}
	}
	if !foundEmitter {
		errors.Add(MissingEmitterError())
	}
	if !foundConsumer {
		errors.Add(MissingConsumerError())
	}
	for _, edge := range v.Graph.Edges {
		if _, ok := allInputs[edge.To]; !ok {
			errors.Add(EdgeInvalidToError(edge))
		}
		allInputs[edge.To] = true

		if _, ok := allOutputs[edge.From]; !ok {
			errors.Add(EdgeInvalidFromError(edge))
		}
		allOutputs[edge.From] = true
	}
	for inputRef, ok := range allInputs {
		if !ok {
			errors.Add(DisconnectedInputError(inputRef))
		}
	}
	for outputRef, ok := range allOutputs {
		if !ok {
			errors.Add(DisconnectedOutputError(outputRef))
		}
	}

	for _, cycle := range v.checkForCycles() {
		errors.Add(CycleError(cycle))
	}

	if !v.isConnected() {
		errors.Add(DisconnectedGraphError())
	}
	return
}

// ValidateActors returns true if the graph actors & roles actually exists
// If not, returns false and a list of ValidationError
func (v *Validator) ValidateActors(actorStorage storage.ActorStorage) ValidationErrorList {
	var errList ValidationErrorList

	for _, node := range v.Graph.Nodes {
		for _, actorID := range node.ActorIDs {
			_, ok := actorStorage.Get(actorID)
			if !ok {
				errList.Add(UnknownActorError(node.ID, actorID.String()))
			}
		}
		for _, actor := range node.Actors {
			id := api.MaybeUUIDFromString(actor)
			if !id.IsZero() {
				if _, ok := actorStorage.Get(id); !ok {
					errList.Add(UnknownActorError(node.ID, actor))
				}
			} else {
				if _, ok := actorStorage.GetByName(actor); !ok {
					errList.Add(UnknownActorError(node.ID, actor))
				}
			}
		}

		var actorKind api.Actor_Kind
		switch node.Type {
		case storage.EmitterNode:
			actorKind = api.Actor_EMITTER
		case storage.ConsumerNode:
			actorKind = api.Actor_CONSUMER
		case storage.WorkerNode:
			actorKind = api.Actor_WORKER
		}

		for _, role := range node.Roles {
			if len(actorStorage.ListRole(actorKind, role)) == 0 {
				errList.Add(NoActorForRoleError(node.ID, role))
			}
		}
	}
	return errList
}

func init() {
	var err error
	idRegexp, err = regexp.Compile(idRegexpS)
	if err != nil {
		panic(err)
	}
}
