package fullenv

import (
	"bufio"
	"fmt"
	"io"
	"os"
	"os/exec"
	"strings"
	"testing"
	"time"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// Fullenv uses the xbus-fullenv binary to coordinate tests
type Fullenv struct {
	configPath    string
	targetDirPath string

	logger xbus.Logger

	cmd   *exec.Cmd
	stdin io.WriteCloser
	err   error

	cmdQuit chan struct{}

	outputChan chan string
}

// NewFullenv instanciate a Fullenv
func NewFullenv(configPath string, logger xbus.Logger) *Fullenv {
	return &Fullenv{
		configPath: configPath,
		logger:     logger,
		outputChan: make(chan string),
	}
}

// NewFullenvForTest returns a fullenv associated to the giveng test
func NewFullenvForTest(t testing.TB, name string, configPath string) *Fullenv {
	logger := GetTestLogger(t)
	fe := NewFullenv(configPath, logger)
	return fe
}

func (fe *Fullenv) initCmd() error {
	cmd := exec.Command("xbus-fullenv", "run", "--no-prompt", fe.configPath)
	stdin, err := cmd.StdinPipe()
	if err != nil {
		return err
	}
	stdout, err := cmd.StdoutPipe()
	if err != nil {
		return err
	}
	stderr, err := cmd.StderrPipe()
	if err != nil {
		return err
	}
	fe.cmd = cmd
	fe.stdin = stdin

	go func() {
		scanner := bufio.NewScanner(stdout)
		for scanner.Scan() {
			s := scanner.Text()
			if strings.HasPrefix(s, "< OK: ") {
				fe.outputChan <- s[6:]
			} else {
				fe.logger.Debugf(s)
			}
		}
		if err := scanner.Err(); err != nil {
			fmt.Fprintln(os.Stderr, "reading stdout:", err)
		}
	}()
	go func() {
		scanner := bufio.NewScanner(stderr)
		for scanner.Scan() {
			s := scanner.Text()
			fe.logger.Noticef(s)
		}
		if err := scanner.Err(); err != nil {
			fmt.Fprintln(os.Stderr, "reading stderr:", err)
		}
	}()
	return nil
}

func (fe *Fullenv) nextOutput() (string, error) {
	select {
	case s := <-fe.outputChan:
		return s, nil
	case <-fe.cmdQuit:
		if fe.err != nil {
			return "", fe.err
		}
		return "", fmt.Errorf("fullenv is closed")
	}
}

func (fe *Fullenv) sendCommand(cmd ...string) (string, error) {
	_, err := fe.stdin.Write([]byte(strings.Join(cmd, " ") + "\n"))
	if err != nil {
		return "", err
	}
	return fe.nextOutput()
}

// Start runs 'xbus-fullenv run' and keep it opened for further interaction
func (fe *Fullenv) Start() error {
	if err := fe.initCmd(); err != nil {
		return err
	}

	fe.cmdQuit = make(chan struct{})

	go func() {
		fe.err = fe.cmd.Run()
		close(fe.cmdQuit)
	}()

	output, err := fe.nextOutput()
	if err != nil {
		return err
	}

	fe.targetDirPath = output

	return nil
}

// Stop stops the fullenv
func (fe *Fullenv) Stop() error {
	_, err := fe.stdin.Write([]byte("quit\n"))
	if err != nil {
		fe.logger.Errorf("%s", err)
	}
	select {
	case <-fe.cmdQuit:
		return fe.Error()
	case <-time.After(2 * time.Second):
		fe.cmd.Process.Kill()
		return fmt.Errorf("Had to kill xbus-fullenv")
	}
}

func (fe *Fullenv) Error() error {
	return fe.err
}

// StartClients sends the 'startup' command to xbus-fullenv
func (fe *Fullenv) StartClients() error {
	_, err := fe.sendCommand("startup")
	return err
}

// ClientConfig returns the configuration path of a given client
func (fe *Fullenv) ClientConfig(name string) (string, error) {
	return fe.sendCommand("client-config", name)
}

// LoadClientConfig load the xbus.COptions of a client
func (fe *Fullenv) LoadClientConfig(name string) (xbus.COptions, error) {
	path, err := fe.ClientConfig(name)
	if err != nil {
		return xbus.COptions{}, err
	}

	return xbus.COptionsFromFile(path, api.Account_ACTOR, fe.logger)
}

// LoadClient loads a client instance
func (fe *Fullenv) LoadClient(name string) (*xbus.Client, error) {
	options, err := fe.LoadClientConfig(name)
	if err != nil {
		return nil, err
	}

	client := xbus.NewClient(options, fe.logger)
	return client, nil
}
